use super::package::Package;
use libarchive::write::{Builder, WriteEntry};
use libarchive::{Entry, WriteFilter, WriteFormat};
use std::collections::HashSet;
use std::fs;
use std::io;
use std::path::{Path, PathBuf};

/// Overarching abstraction that orchestrates updating the repositories stored on the server
pub struct RepoGroupManager {
    repo_dir: PathBuf,
    pkg_dir: PathBuf,
    default_arch: String,
}

impl RepoGroupManager {
    pub fn new<P1: AsRef<Path>, P2: AsRef<Path>>(
        repo_dir: P1,
        pkg_dir: P2,
        default_arch: &str,
    ) -> Self {
        RepoGroupManager {
            repo_dir: repo_dir.as_ref().to_path_buf(),
            pkg_dir: pkg_dir.as_ref().to_path_buf(),
            default_arch: String::from(default_arch),
        }
    }

    pub fn sync(&mut self, repo: &str, arch: &str) -> io::Result<()> {
        let subrepo_path = self.repo_dir.join(repo).join(arch);

        let mut ar_db = Builder::new();
        ar_db.add_filter(WriteFilter::Gzip)?;
        ar_db.set_format(WriteFormat::PaxRestricted)?;

        let mut ar_files = Builder::new();
        ar_files.add_filter(WriteFilter::Gzip)?;
        ar_files.set_format(WriteFormat::PaxRestricted)?;

        let mut ar_db = ar_db.open_file(subrepo_path.join(format!("{}.tar.gz", repo)))?;
        let mut ar_files =
            ar_files.open_file(subrepo_path.join(format!("{}.files.tar.gz", repo)))?;

        for entry in subrepo_path.read_dir()? {
            let entry = entry?;

            if entry.file_type()?.is_dir() {
                // The desc file needs to be added to both archives
                let path_in_tar = PathBuf::from(entry.file_name()).join("desc");
                let src_path = entry.path().join("desc");

                let mut ar_entry = WriteEntry::new();
                ar_entry.set_pathname(&path_in_tar);
                ar_entry.set_mode(0o100644);

                ar_db.append_path(&mut ar_entry, &src_path)?;
                ar_files.append_path(&mut ar_entry, src_path)?;

                // The files file is only required in the files database
                let path_in_tar = PathBuf::from(entry.file_name()).join("files");
                let src_path = entry.path().join("files");

                let mut ar_entry = WriteEntry::new();
                ar_entry.set_pathname(&path_in_tar);
                ar_entry.set_mode(0o100644);

                ar_files.append_path(&mut ar_entry, src_path)?;
            }
        }

        ar_db.close().and(ar_files.close()).map_err(Into::into)
    }

    pub fn add_pkg_from_path<P: AsRef<Path>>(&mut self, repo: &str, path: P) -> io::Result<()> {
        let mut pkg = Package::open(&path)?;
        pkg.calculate_checksum()?;

        let archs = self.add_pkg_in_repo(repo, &pkg)?;

        // We add the package to each architecture it was added to by hard-linking the provided
        // package file. This prevents storing a package of type "any" multiple times on disk.
        for arch in archs {
            let arch_repo_pkg_path = self.pkg_dir.join(repo).join(arch);
            let dest_pkg_path = arch_repo_pkg_path.join(pkg.file_name());

            fs::create_dir_all(&arch_repo_pkg_path)?;
            fs::hard_link(&path, dest_pkg_path)?;
        }

        fs::remove_file(path)
    }

    /// Add a package to the given repo, returning to what architectures the package was added.
    pub fn add_pkg_in_repo(&mut self, repo: &str, pkg: &Package) -> io::Result<HashSet<String>> {
        let mut arch_repos: HashSet<String> = HashSet::new();

        if pkg.info.arch != "any" {
            self.add_pkg_in_arch_repo(repo, &pkg.info.arch, pkg)?;
            arch_repos.insert(pkg.info.arch.clone());
        }
        // Packages of arch "any" are added to every existing arch
        else {
            arch_repos.insert(self.default_arch.clone());

            let repo_dir = self.repo_dir.join(repo);

            if repo_dir.exists() {
                for entry in repo_dir.read_dir()? {
                    arch_repos.insert(entry?.file_name().to_string_lossy().to_string());
                }
            }

            for arch in arch_repos.iter() {
                self.add_pkg_in_arch_repo(repo, arch, pkg)?;
            }
        }

        Ok(arch_repos)
    }

    pub fn add_pkg_in_arch_repo(
        &mut self,
        repo: &str,
        arch: &str,
        pkg: &Package,
    ) -> io::Result<()> {
        let pkg_dir = self
            .repo_dir
            .join(repo)
            .join(arch)
            .join(format!("{}-{}", pkg.info.name, pkg.info.version));

        // We first remove the previous version of the package, if present
        self.remove_pkg_from_arch_repo(repo, arch, &pkg.info.name)?;

        fs::create_dir_all(&pkg_dir)?;

        let mut desc_file = fs::File::create(pkg_dir.join("desc"))?;
        pkg.write_desc(&mut desc_file)?;

        let mut files_file = fs::File::create(pkg_dir.join("files"))?;
        pkg.write_files(&mut files_file)?;

        self.sync(repo, arch)
    }

    pub fn remove_repo(&mut self, repo: &str) -> io::Result<bool> {
        let repo_dir = self.repo_dir.join(&repo);

        if !repo_dir.exists() {
            Ok(false)
        } else {
            fs::remove_dir_all(&repo_dir).and(fs::remove_dir_all(self.pkg_dir.join(repo)))?;

            Ok(true)
        }
    }

    pub fn remove_arch_repo(&mut self, repo: &str, arch: &str) -> io::Result<bool> {
        let sub_path = PathBuf::from(repo).join(arch);
        let repo_dir = self.repo_dir.join(&repo).join(&sub_path);

        if !repo_dir.exists() {
            Ok(false)
        } else {
            fs::remove_dir_all(&repo_dir).and(fs::remove_dir_all(self.pkg_dir.join(sub_path)))?;

            Ok(true)
        }
    }

    pub fn remove_pkg_from_arch_repo(
        &mut self,
        repo: &str,
        arch: &str,
        pkg_name: &str,
    ) -> io::Result<bool> {
        let arch_repo_dir = self.repo_dir.join(repo).join(arch);

        if !arch_repo_dir.exists() {
            return Ok(false);
        }

        for entry in arch_repo_dir.read_dir()? {
            let entry = entry?;

            // Make sure we skip the archive files
            if !entry.metadata()?.is_dir() {
                continue;
            }

            let file_name = entry.file_name();
            let file_name = file_name.to_string_lossy();

            // The directory name should only contain the name of the package. The last two parts
            // when splitting on a dash are the pkgver and pkgrel, so we trim those
            let name_parts = file_name.split('-').collect::<Vec<_>>();
            let name = name_parts[..name_parts.len() - 2].join("-");

            if name == pkg_name {
                fs::remove_dir_all(entry.path())?;

                // Also remove the old package archive
                let arch_repo_pkg_dir = self.pkg_dir.join(repo).join(arch);

                arch_repo_pkg_dir.read_dir()?.try_for_each(|res| {
                    res.and_then(|entry: fs::DirEntry| {
                        let file_name = entry.file_name();
                        let file_name = file_name.to_string_lossy();

                        // Same trick, but for package files, we also need to trim the arch
                        let name_parts = file_name.split('-').collect::<Vec<_>>();
                        let name = name_parts[..name_parts.len() - 3].join("-");

                        if name == pkg_name {
                            fs::remove_file(entry.path())
                        } else {
                            Ok(())
                        }
                    })
                })?;

                return Ok(true);
            }
        }

        Ok(false)
    }
}
